import { fn, URI, symbol } from '@traversable/registry';
import { t } from '@traversable/schema';
import { isRequired, property } from './properties.js';
import * as JsonSchema from './jsonSchema.js';
/** @internal */
const phantom = () => void 0;
/** @internal */
function handleOptionality(x) {
  let {
    [symbol.optional]: ix,
    ...y
  } = x;
  y = Recursive.fromJsonSchema(y);
  if (typeof ix !== 'number') return y;else while (ix-- > 0) y = t.optional(y);
  return y;
}
/** @internal */
function recoverTupleOptionality(xs, {
  min,
  max
}) {
  if (min !== max) {
    const firstOptional = min;
    const req = xs.slice(0, firstOptional);
    const opt = xs.slice(firstOptional).map(t.optional.def);
    return [...req, ...opt];
  } else return xs;
}
export var Recursive;
(function (Recursive) {
  Recursive.toJsonSchema = x => {
    switch (true) {
      default:
        return fn.exhaustive(x);
      case x.tag === URI.symbol:
        return phantom;
      case x.tag === URI.bigint:
        return phantom;
      case x.tag === URI.never:
        return phantom;
      case x.tag === URI.undefined:
        return phantom;
      case x.tag === URI.void:
        return phantom;
      case x.tag === URI.any:
        return () => JsonSchema.RAW.any;
      case x.tag === URI.unknown:
        return () => JsonSchema.RAW.any;
      case x.tag === URI.null:
        return () => JsonSchema.RAW.null;
      case x.tag === URI.boolean:
        return () => JsonSchema.RAW.boolean;
      case x.tag === URI.integer:
        return JsonSchema.IntegerSchema.toJsonSchema;
      case x.tag === URI.number:
        return JsonSchema.NumberSchema.toJsonSchema;
      case x.tag === URI.string:
        return JsonSchema.StringSchema.toJsonSchema;
      case x.tag === URI.optional:
        return JsonSchema.OptionalSchema(x.def()).toJsonSchema;
      case x.tag === URI.eq:
        return JsonSchema.EqSchema(x.def).toJsonSchema;
      case x.tag === URI.array:
        return JsonSchema.ArraySchema(x.def()).toJsonSchema;
      case x.tag === URI.record:
        return JsonSchema.RecordSchema(x.def()).toJsonSchema;
      case x.tag === URI.union:
        return JsonSchema.UnionSchema(fn.map(x.def, v => v())).toJsonSchema;
      case x.tag === URI.intersect:
        return JsonSchema.IntersectSchema(fn.map(x.def, v => v())).toJsonSchema;
      case x.tag === URI.tuple:
        return JsonSchema.TupleSchema(fn.map(x.def, v => v())).toJsonSchema;
      case x.tag === URI.object:
        {
          const required = Object.keys(x.def).filter(isRequired(x.def));
          return () => ({
            type: 'object',
            required,
            properties: fn.map(x.def, (v, k) => property(required)(v(), k))
          });
        }
    }
  };
  Recursive.fromJsonSchema = x => {
    switch (true) {
      default:
        return fn.exhaustive(x);
      case symbol.optional in x:
        return handleOptionality(x);
      case JsonSchema.is.null(x):
        return t.null;
      case JsonSchema.is.boolean(x):
        return t.boolean;
      case JsonSchema.is.integer(x):
        return t.integer;
      case JsonSchema.is.number(x):
        return t.number;
      case JsonSchema.is.string(x):
        return t.string;
      case JsonSchema.is.any(x):
        return t.unknown;
      case JsonSchema.is.record(x):
        return t.record(x.additionalProperties);
      case JsonSchema.is.union(x):
        return t.union.def(x.anyOf);
      case JsonSchema.is.enum(x):
        return t.union.def(x.enum.map(_ => t.eq(_)));
      case JsonSchema.is.const(x):
        return t.eq.def(x.const);
      case JsonSchema.is.intersect(x):
        return t.intersect.def(x.allOf);
      case JsonSchema.is.tuple(x):
        return t.tuple.def(recoverTupleOptionality(x.items, {
          min: x.minItems,
          max: x.maxItems
        }));
      case JsonSchema.is.array(x):
        return t.array.def(x.items);
      case JsonSchema.is.object(x):
        return t.object(fn.map(x.properties, (v, k) => x.required.includes(`${k}`) ? v : t.optional(v)));
    }
  };
})(Recursive || (Recursive = {}));
/**
 * ## {@link fromJsonSchema `fromJsonSchema`}
 *
 * Recursively convert a JSON schema into a traversable schema.
 *
 * __Roundtrip property:__
 *
 * Thoroughly tested using randomly generated schemas to be invertible when paired
 * with {@link toJsonSchema} without any loss of information.
 */
export const fromJsonSchema = /*#__PURE__*/JsonSchema.fold(Recursive.fromJsonSchema);
/**
 * ## {@link toJsonSchema `toJsonSchema`}
 *
 * Recursively convert a traversable schema into its JSON schema representation.
 *
 * __Roundtrip property:__
 *
 * Thoroughly tested using randomly generated schemas to be invertible with
 * {@link fromJsonSchema `fromJsonSchema`} without any loss of information.
 *
 * __Note:__ If you find yourself reaching for {@link toJsonSchema `toJsonSchema`},
 * keep in mind that by installing the `@traversable/schema-to-json-schema` package,
 * all of the built-in schemas from the `@traversable/schema` package come
 * equipped with a `.toJsonSchema` getter method.
 *
 * Prefer using the `.toJsonSchema` method if possible -- not only is it
 * easier to use, but it's also more performant, since the implementation
 * is not recursive.
 */
export const toJsonSchema = /*#__PURE__*/t.fold(Recursive.toJsonSchema);
//# sourceMappingURL=recursive.js.map